;; Generated package description from doom-game.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "doom-game" "0.1.0.20260818.4" "DOOM on Emacs" '((emacs "31.0")) :commit "c01017027e9cb8a156409a3f2ea2a647ea501205" :authors '(("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '("Daniel Mendler" . "mail@daniel-mendler.de") :keywords '("games") :url "https://github.com/minad/doom-on-emacs")
